/*
   Copyright 2010 Alfredo Braunstein, Farbod Kayhan and Riccardo Zecchina

   This file is a part of a lossy encoder over GF(q).

   This is a free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdio>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <set>
#include "formula.hpp"
#include "random.hpp"
#include "gflib.hpp"

using namespace std;

extern int Q;
extern int q;
extern int N;

extern int M;
extern vector<int> randseq;
extern vector<int> codeword;
extern double L;

extern vector<varstruct> variables;
extern vector<clausestruct> clauses;

vector<int> info_vars;
vector<int> diagonal;
vector<int> diagonal_check;

void readformula(char const * filename) 
{
	ifstream fg(filename);

	if (!fg) {
		cout<<"Error in Reading Formula: could not open file!"<<endl;
		exit(-1);
	}

//reading the header line Q, N, M
	fg >> Q >> N >> M;

	cout << "#variables: " << N << "\t" << "#checks: " << M << "\t" << "GF: " << Q << "\t";

	q = log(Q) / log(2) + 0.1;
	cout << "q = Log2(Q) = " << q << endl;

	clauses.clear();
	variables = vector<varstruct>(N, varstruct(Q));
	randseq = vector<int>(N);
	codeword = vector<int>(N);

	int var;
	int m = 0;

	clausestruct c;
	while (fg >> var) {
		if (var < 0) {
			clauses.push_back(c);
			c = clausestruct();
			++m;
			continue;
		}

		c.varlist.push_back(var);
		c.position.push_back(variables[var].clauselist.size());
		variables[var].messages.push_back(DMes(Q, 1.));
		variables[var].clauselist.push_back(m);
		int w;
		fg >> w;
		c.weightlist.push_back(w);
	}
	assert(m == M);
}

void leaf_removal()
{
	set<int> leaves;
	vector<int> deg(N);
	vector<int> active_check(M, 1);
	for (int i = 0; i < N; ++i) {
		int d = variables[i].vardeg();
		deg[i] = d;
		if (d == 1)
			leaves.insert(i);
	}
	while (!leaves.empty()) {
		int var = *leaves.begin();
		leaves.erase(leaves.begin());
		assert(deg[var] == 1);
		deg[var] = 0;
		diagonal.push_back(var);
		for (int j = 0; j < variables[var].vardeg(); ++j) {
			int c = variables[var].clauselist[j];
			if (active_check[c]) {
				diagonal_check.push_back(c);
				active_check[c] = 0;
				for (int k = 0; k < clauses[c].clausedeg(); ++k) {
					int var2 = clauses[c].varlist[k];
					if (deg[var2] == 1) {
						info_vars.push_back(var2);
						leaves.erase(var2);
					} else if (deg[var2] == 2)
						leaves.insert(var2);
					deg[var2]--;
				}
			}
		}
	}
	cout << "diagonal (" << diagonal.size() << ") :";
	for (int i = diagonal.size(); i--; )
		cout << diagonal[i] << " ";
	cout << endl;	
	cout << "info bits (" << info_vars.size() << ") :";

	for (int i = info_vars.size(); i--; )
		cout << info_vars[i] << " ";

	cout << endl;	
	assert(int(info_vars.size() + diagonal.size()) == N);
}



void createseq()
{
	for (int i = 0; i < N; ++i)
		randseq[i] = intrand(Q);
}

// binary distance (Hamming Distance) of two symbols in GF(Q)
int gfdis (int a, int b)
{
	int dist = 0;
	int z = a ^ b;
	while (z) {
		dist += z & 1;
		z >>= 1;
	}
	return dist;
}
// binary distance between two sequences of length N in GF(Q)
int gfdis_seq(std::vector<int> const & a, std::vector<int> const & b)
{
	int dist_seq = 0;
	for(int i = 0; i < N; ++i)
		dist_seq += gfdis(a[i],b[i]);

	return dist_seq;
}

// binary distortion between two sequences of lenght N in GF(Q)
double gfdistortion(std::vector<int> const & a, std::vector<int> const & b)
{
	return gfdis_seq(a, b) / (double(N * q));
}




void initialmemo()
{
	real_t norm = 0;

	for (int k = 0; k < Q; ++k)
		norm += exp(-L * gfdis(0, k));

//fixing the external field given the randseq---------------------------
	for (int i = 0; i < N; ++i) {
		for (int j = 0; j < variables[i].vardeg(); ++j) {
			for (int k = 0; k < Q; ++k) 
				variables[i].messages[j][k] = 1.;
		}
		for (int k = 0; k < Q; ++k) {
			variables[i].exfield[k] = exp(-L * gfdis(randseq[i], k)) / norm;
			variables[i].marginal[k] = 1;
		}
	}
}

