/*
   Copyright 2010 Alfredo Braunstein, Farbod Kayhan and Riccardo Zecchina

   This file is a part of GFRBP, a lossy compressor of random independent
   binary sources in GF(q).

   This is a free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "random.hpp"
#include "formula.hpp"
#include "cavity.hpp"
#include "gflib.hpp"

#include <cassert>
#include <vector>
#include <algorithm>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

using namespace std;

extern double gamma1;               //The reinforcement parameter gamma_1
extern double gamma0;               //The reinforcement parameter gamma_0
extern double gamma_fun;
extern double L;
extern int N;
extern int M;
extern int Q;
extern int q;


extern double epsilon;              //The precision for convergence
extern int l_max;                   //Max number of iterations
int conv_itnum;			    //Number of iteration after convergence
int min_energy;        		    //Minimum Energy reached by RBP


vector<int> randseq;
vector<int> codeword;
vector<int> MEseq;

extern vector<varstruct> variables;
extern vector<clausestruct> clauses;


extern vector<int> info_vars;
extern vector<int> diagonal;
extern vector<int> diagonal_check;




namespace {

#define FAST_POW

#ifdef FAST_POW 
#	define pow fast_pow
#endif

double fast_pow(double a, double b) {
	int const tmp = (*(1 + (int *)&a));
	int const tmp2 = (int)(b * (tmp - 1072632447) + 1072632447);
	double p = 0.0;
	*(1 + (int * )&p) = tmp2;
	return p;
}

void FTold(DMes & m)
{
	static DMes fm(Q);
	for (int u = Q; u >>= 1; ) {
		for (int i = Q; i--; ) 
			fm[i] = m[i ^ u] + ((i & u) ? -m[i] : m[i]);
		swap(fm, m);
	}
}

void FT(DMes & m)
{
	for (int u = Q; u >>= 1; ) {
		int const t = u - 1;
		for (int j = Q >> 1; j--; ) {
			int const x = j & t;
			int const i = ((j ^ x) << 1) ^ x; 
			real_t & miu = m[i ^ u];
			real_t const oldmiu = miu;
			real_t & mi = m[i];
			miu  = -miu + mi;
			mi += oldmiu;
		}
	}
}

DMes multiply(DMes const & a, DMes const & b)
{
	DMes c(Q);
	for (int k = 0; k < Q; ++k)
		c[k] = a[k] * b[k];
	return c;
}

void normalize(real_t * a)
{
	real_t * const nb = a + Q;
	real_t norm = 0;
	for (real_t * b = a; b < nb; ++b) {
		if (*b < 0)
			*b = 0;
		else
			norm += *b;
	}
	//assert(norm > 0);
	if (norm <= 1e-30) {
		for (real_t * b = a; b < nb; ++b) 
			*b = 1 / Q;
		return;
	}
	real_t const invnorm = 1. / norm;
	for (real_t * b = a; b < nb; ++b)
		*b *= invnorm;
}

void normalize(DMes & m)
{
	normalize(&m[0]);
}


} // anon namespace


double gfupdate()
{
	static GF calc(q);
	DMes aux(Q);

//initializing the permutation vector
	std::vector<int> permutation(M);
	for (int i = 0; i < M; ++i)
		permutation[i] = i;
	random_shuffle(permutation.begin(), permutation.end());

	real_t diff = 0;

	for (int ii = 0; ii < M; ++ii) {
		int i = permutation[ii];
		int const di = clauses[i].clausedeg();
		vector<DMes> hh(di, DMes(Q));

		for (int j = 0; j < di; ++j) {

			int i2 = clauses[i].varlist[j];
			int j2 = clauses[i].position[j];
			assert(variables[i2].clauselist[j2] == i);

			for (int k = 0; k < Q; ++k)
				aux[k] = variables[i2].exfield[k] * variables[i2].marginal[k];
			assert(variables[i2].vardeg() <= 2);
			for (int l = 0; l < variables[i2].vardeg(); ++l)
				if (l != j2)
					for (int k = 0; k < Q; ++k)
						aux[k] *= variables[i2].messages[l][k];
			normalize(aux);

			int const c = clauses[i].weightlist[j];
			for (int k = 0; k < Q; ++k) 
				hh[j][k] = aux[calc.div(k,c)];

			FT(hh[j]);
		}
	
	
		if (di <= 4) {	
			for (int j = 0; j < di; ++j) {
				DMes m(Q, 1.);
				for (int jj = 0; jj < di; ++jj) if (j != jj)
					for (int k = 0; k < Q; ++k)
						m[k] *= hh[jj][k];
				FT(m);
				normalize(&m[0]);
				int i2 = clauses[i].varlist[j];
				int j2 = clauses[i].position[j];
				for (int k = 0; k < Q; ++k) {
					real_t & oldmes = variables[i2].messages[j2][k];
					int c = clauses[i].weightlist[j];
					int pik = calc.mult(k, c);
					real_t newmes = m[pik];
					diff = max(diff, fabs(newmes - oldmes));
					oldmes = newmes;
				}
			}
		} else {
			Cavity<DMes> C(hh, DMes(Q, 1.), multiply);
			Cavity<DMes>::iterator cit = C.begin();
			for (int j = 0; j < di; ++j, ++cit) {
				DMes m = *cit;
				FT(m);
				normalize(&m[0]);
				int i2 = clauses[i].varlist[j];
				int j2 = clauses[i].position[j];
				for (int k = 0; k < Q; ++k) {
					real_t & oldmes = variables[i2].messages[j2][k];
					int c = clauses[i].weightlist[j];
					int pik = calc.mult(k, c);
					real_t newmes = m[pik];
					diff = max(diff, fabs(newmes - oldmes));
					oldmes = newmes;
				}
			}
		}
	}


//calculating the new marginals
	for (int i = 0; i < N; ++i) {
		for (int k = 0; k < Q; ++k)
			variables[i].marginal[k] = variables[i].exfield[k] * variables[i].marginal[k];
		for (int j = 0; j < variables[i].vardeg(); ++j)
			for (int k = 0; k < Q; ++k)
				variables[i].marginal[k] *= variables[i].messages[j][k];
		
		for (int k = 0; k < Q; ++k)
			variables[i].marginal[k] = pow(variables[i].marginal[k], gamma_fun);
		normalize(variables[i].marginal);
	}
	return diff;
}


void get_codeword()
{
	static GF calc(q);
	codeword = vector<int>(N, -1);
	for (int l = 0; l < int(info_vars.size()); ++l) {
		int i = info_vars[l];
		int hard_ind = 0;
		double hard = variables[i].marginal[0];
		for (int k = 0; k < Q; k++){
			if (variables[i].marginal[k] > hard){
				hard_ind = k;
				hard = variables[i].marginal[k];
			}
		}	
		codeword[i] = hard_ind;
	}

	for (int l = diagonal.size(); l--; ) {
		int i = diagonal[l];
		int c = diagonal_check[l];
		int sum = 0;
		int coeff = 0;
		for (int j = 0; j < clauses[c].clausedeg(); ++j) {
			int w = clauses[c].weightlist[j];
			int v = clauses[c].varlist[j];
			if (v == i)
				coeff = w;
			else {
				assert(codeword[v] >= 0);
				sum ^= calc.mult(w, codeword[v]);
			}
		}
		codeword[i] = calc.div(sum, coeff);
	}
}

//hard decision and energy calculation
int energy_cal (void)
{
	static GF calc(q);
//hard decision as an attemp to obtain a codeword!
	vector<int> decision(N);

	for (int i = 0; i < N; ++i){
		int hard_ind = 0;
		double hard = variables[i].marginal[0];
		for (int k = 0; k < Q; k++){
			if (variables[i].marginal[k] > hard){
				hard_ind = k;
				hard = variables[i].marginal[k];
			}
		}
		decision[i] = hard_ind;
	}

//calculating the number of checks which are not satisfied (energy)
	int unsat_num = 0;
	for (int i=0; i<M; i++){
		int sum_M = 0;
		for(int j=0;  j<clauses[i].clausedeg(); ++j) {
			sum_M ^= calc.mult(decision[clauses[i].varlist[j]],clauses[i].weightlist[j]);
		}
		unsat_num += (sum_M != 0);
	}

	return unsat_num;
}


real_t gfreinforced()
{
	int iter_num = 1;       //iteration counter
	int energy = M;         //energy of a given marginals after hard decision
	conv_itnum = l_max;
	min_energy = M;
	cout << "converge:" << endl;

	MEseq = vector<int>(N);

	double err;
	real_t mindist = 1;
	do {
		//gamma_fun = 1 - gamma0 * pow(gamma1, iter_num);
//linear gamma update is also possible, i.e. 
		//gamma_fun = gamma0 + iter_num * iter_num * gamma1;
		real_t x = iter_num / 100. ;
		gamma_fun = 0.08 * x * x / (1 + x * x);

		err = gfupdate();
		++iter_num;

		energy = energy_cal();
		get_codeword();
		real_t dist = gfdistortion(randseq, codeword);
		if (dist < mindist) {
			mindist = dist;
			MEseq = codeword;
		}
		cout << "it:" << setw(5) << iter_num << " gamma:" << setw(8) << gamma_fun << " energy: " << setw(5) << right << energy << " dist: " << setw(8) << left << dist  << " mindist: " << setw(9) << left << mindist << endl;
	} while (iter_num < l_max  && err > epsilon && energy > 0);

	conv_itnum = iter_num;
	if (err < epsilon) 
		cout << "converged" << endl;
	if (energy == 0) 
		cout << "A solution is found after iteration: " << iter_num << endl;
	cout << "energy= " << energy << endl;
	cout << "min_energy= " << min_energy << endl;
	cout << "min_dist= " << mindist << endl;
	return mindist;
}


real_t gfdistortion()
{
	return gfdistortion(randseq, MEseq);
}

void writesolution(int simul_num)
{
	ostringstream op;
	op << "output/N" << N << "_M" << M << "_g" 
		<< gamma0 << "_G" << gamma1 << "_s" << simul_num << flush;
	ofstream Svector ((op.str() + "_SS.txt").c_str());
	for (int i = 0; i < N; ++i)
		Svector << i << " S= " << randseq[i] 
		<< " \t MEcseq=" << MEseq[i] << endl;
	
//measuring the distance of S and the found codeword
	cout << "Distortion: " << gfdistortion(randseq, MEseq) << endl;

	ofstream meditfile ((op.str() + "_MED.txt").c_str());
	meditfile << "Distortion = " << gfdistortion(randseq, MEseq) 
		<< " min energy = " << min_energy 
		<< " #it= " << conv_itnum << endl;

}



